import pandas as pd
import warnings


class Dataframes:
    def read_spreadsheet(self, file, return_ext=False):

        # Get file MIME
        file_mime = file.content_type

        # Read CSV
        if file_mime == 'text/csv':
            file_ext = 'csv'

            # Read in data
            input_data = pd.read_csv(file)
            
        # Read TSV
        elif file_mime == 'text/tab-separated-values':
            file_ext = 'tsv'

            # Read in data
            input_data = pd.read_csv(file, sep='\t')
            
        # Read Excel
        elif (file_mime == 'application/vnd.ms-excel' or
            file_mime == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'):
            file_ext = 'xlsx'

            # Read in data
            input_data = pd.read_excel(file)

        # Unsupported filetype
        else:
            warnings.warn(f'"{file_mime}" is not a supported filetype.')


        return input_data, file_ext if return_ext else input_data

    def drop_val(self, df, col_name, value):  # *
        """Drops row from df where column entry is equal to value.

        Args:
            df (pandas DataFrame): DataFrame containing rows to drop.
            col_name (str): Name of column to investigate.
            value (pandas object): Value to drop instances of.

        Returns:
            pandas DataFrame: DataFrame with rows dropped.
        """
        return self.pull_not_val(df, col_name, value).reset_index(drop=True)

    def pull_val(self, df, col_name, value):  # *
        """Retrieves rows from df where column entry is equal to value.

        Args:
            df (pandas DataFrame): DataFrame containing rows to pull.
            col_name (str): Name of column to investigate.
            value (pandas object): Value to pull instances of.

        Returns:
            pandas DataFrame: DataFrame where value is found.
        """
        return df.where(df[col_name] == value).dropna()

    def pull_not_val(self, df, col_name, value):  # *
        """Retrieves rows from df where column entry is not equal to value.

        Args:
            df (pandas DataFrame): DataFrame containing rows to pull.
            col_name (str): Name of column to investigate.
            value (pandas object): Value to not pull instances of.

        Returns:
            pandas DataFrame: DataFrame where value is not found.
        """
        return df.where(df[col_name] != value).dropna()


import pandas as pd
import numpy as np


class Duplicates:
    def filter(self, df, subsets, return_duplicates=False, how='all'):  # *
        """Filters duplicates to either return DataFrame subset of duplicates or non-duplicates.

        Args:
            df (pandas DataFrame): DataFrame to perform duplicate search on.
            subsets (list): List of column names to compare in duplicate search.
                            All must match to be considered a duplicate.
            return_duplicates (bool, optional): If True, returns non-duplicate df else returns duplicate df.
                                                Defaults to False.
            how (str): 

        Returns:
            pandas DataFrame: Filtered df.
        """
        if how == 'any' and len(subsets) > 1:
            # Checks for occurances in ANY column of 'subsets'
            index_arr = np.array([])
            for col in subsets:     
                index_arr = np.concatenate((index_arr, indices(df, col)))
            
            index_arr = np.unique(index_arr)
            
            return df.loc[df.index.isin(index_arr)] if return_duplicates else df.loc[~df.index.isin(index_arr)]
        
        elif how == 'all':
            # Duplicates in ALL columns of 'subsets' must agree
            duplicates = pd.DataFrame.duplicated(df, subset=subsets)
            
        else:
            # Raise error of how is not 'any' or 'all'
            raise ValueError('how argument must be either \'any\' or \'all\'')
        
        return df[duplicates] if return_duplicates else df[~duplicates]

    def remove(self, df, subsets):  # *
        """Removes duplicates from df. Keeps first occurance in df.

        Args:
            df (pandas DataFrame): DataFrame to perform dupicate search on.
            subsets (list): List of column names to compare in duplicate search.
                            All must match to be considered a duplicate.

        Returns:
            pandas DataFrame: DataFrame with duplicates removed.
        """
        return self.filter(df, subsets=subsets)

    def indices(self, df, column):  # *
        return list(df[pd.DataFrame.duplicated(df, subset=[column])].index)

    def average(self, df, subsets, average_by):  # *
        """Averages duplicates from df.

        Args:
            df (pandas DataFrame): DataFrame to perform duplicate search on.
            subsets (list): List of column names to compare in duplicate search. 
                            All must match to be considered a duplicate.
            average_by (str): Column name to average along.

        Returns:
            pandas DataFrame: DataFrame with duplicates averaged.
        """
        df_copy = df.copy()
        
        # Filter out duplicates for merge df
        out = self.filter(df_copy, subsets)
        orig_index = out.index
            
        # Ensure average_by exists in the dataframe
        if average_by not in df.columns:
            raise ValueError('average_by column not found in DataFrame')
        
        # Cast value column to float (for avg computation)
        try:
            df_copy[average_by] = df_copy[average_by].astype(float)
        except ValueError:
            raise ValueError('average_by column does not contain float castable values')
        
        # Group duplicate rows and aggregate by mean (adds mean to end of value column)
        avg_value = df_copy.groupby(by=subsets).agg({average_by: ['mean']})
        avg_value.columns = [col[0] for col in avg_value.columns.values]

        # Drop original value column ('value' --> 'valuemean')
        out = out.drop(columns=[average_by])  # drop original values

        # Add averaged values to dataframe and preserve original index
        out = pd.merge(out, avg_value, on=subsets).set_index(orig_index)
        out = out[df_copy.columns]
        
        return out
